/**
 * Copyright (C) 2003 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package org.objectweb.perseus.pool.lib;

import org.objectweb.perseus.pool.api.Pool;
import org.objectweb.perseus.pool.api.PoolAttributes;
import org.objectweb.perseus.pool.api.PoolMatchFactory;
import org.objectweb.perseus.pool.api.PoolException;
import org.objectweb.perseus.dependency.api.DependencyGraph;
import org.objectweb.perseus.dependency.api.DeadLockException;
import org.objectweb.fractal.api.control.BindingController;
import org.objectweb.util.monolog.api.Logger;
import org.objectweb.util.monolog.api.BasicLevel;

import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.Map;

/**
 * The class <b>LArrayPool</b> implements a Pool as an array of PoolResource,
 * managing free/active resources through two ArrayList objects. An unlimited
 * size and a timeout features are supported. This implementation is thread
 * safe. 
 * This implementation is able to manage an InactiveTTL for the free 
 * pooled resources, but by default the resource has not a TTL (negative value).
 * This implementation is able to manage a TTL for each  
 * pooled resources, but by default the resource has not a TTL (negative value).
 *
 * @author S.Chassande-Barrioz
 */
public class ArrayListPool
		implements Pool, PoolAttributes, BindingController {

	public final static String POOL_MATCH_FACTORY_BINDING = "pool-match-factory";
	public final static String DEPENDENCY_GRAPH_BINDING = "dependency-graph";

    /**
     * The int value used for the timeout or the max size
     */
    public final static int UNLIMITED = -1;

    /**
     * The default minimum size of the pool: 0
     */
    public final static int DEFAULT_MIN_SIZE = 0;

    /**
     * The default maximum size of the pool: UNLIMITED
     */
    public final static int DEFAULT_MAX_SIZE = UNLIMITED;

    /**
     * The default time out for a request: 100ms
     */
    public final static int DEFAULT_TIME_OUT = 300;

    private PoolMatchFactory matchFactory = null;
    private long freettl = 0;
    private long ttl = 0;
    private long timeOut = DEFAULT_TIME_OUT;
    private int minSize = DEFAULT_MIN_SIZE;
    private int maxSize = DEFAULT_MAX_SIZE;
    private List free = new ArrayList();
    private List timesOfFree = new ArrayList();
    private HashMap res2ttl = new HashMap();
    private List used = new ArrayList();
    private List users = new ArrayList();
	private boolean started = false;
	private Logger logger = null;
	private DependencyGraph dg = null;


	/**
	 * destroy the resource that the ttl or freettl is expired
	 */
	private void cleanExpiredResource() {
		long time = 0;
		if (freettl > 0 || ttl > 0) {
			time = Calendar.getInstance().getTimeInMillis();
			for(int i=0; i<free.size();) {
				boolean remove = false;//indicates if the resource has to be removed
		        Object resource = free.get(i);
			    if (freettl > 0) {
				    Long l = (Long) timesOfFree.get(i);
				    if (l == null) {
				        throw new IllegalThreadStateException("The resource '" + resource + "' has no inactivettl !");
				    }
			        remove = (time - l.longValue()) > freettl;
					if (remove && logger != null && logger.isLoggable(BasicLevel.INFO)) {
					    logger.log(BasicLevel.INFO, "Resource is being destroyed (too long time unused): '" + resource + "'.");
					}
			    }
			    if (!remove && ttl > 0) {
				    //inactivettll does not expire ==> check the ttl
				    Long l = (Long) res2ttl.get(resource);
				    if (l == null) {
				        throw new IllegalThreadStateException("The resource '" + resource + "' has no ttl !");
				    }
				    remove = (time - l.longValue()) > ttl;
					if (remove && logger != null && logger.isLoggable(BasicLevel.INFO)) {
					    logger.log(BasicLevel.INFO, "Resource is being destroyed (too old): '" + resource + "'.");
					}
				}
				if (remove) {
				    //one of both ttl has expired ==> remove the resource
					free.remove(i);
				    if (freettl > 0) {
			            timesOfFree.remove(i);
				    }
					if (ttl>0) {
					    res2ttl.remove(resource);
					}
					matchFactory.destroyResource(resource);
				} else {
					i++;
				}
			}
		}
	}

	private void updateToMinsize(Object hints) throws PoolException {
		cleanExpiredResource();
		int delta = minSize - free.size() - used.size();
		if (matchFactory != null && delta>0) {
			for(int i=0; i<delta; i++) {
			    Object resource = createResource(hints);
				free.add(resource);
				if (freettl > 0) {
					timesOfFree.add(new Long(Calendar.getInstance().getTimeInMillis()));
				}

			}
		}
	}

	private Object createResource(Object hints) throws PoolException {
		Object resource = matchFactory.createResource(hints);
		if (resource == null) {
			if (started) {
				throw new PoolException("The pool resource factory does not create a new resource with the hints: " + hints);
			} else if (logger != null) {
				logger.log(BasicLevel.WARN, "The pool resource factory does not create a new resource with the hints: " + hints);
			}
		} else {
	        long t = 0;
		    if (ttl > 0) {
		        t = Calendar.getInstance().getTimeInMillis();
		        res2ttl.put(resource, new Long(t));
		    }
			if (logger != null && logger.isLoggable(BasicLevel.INFO)) {
			    logger.log(BasicLevel.INFO, "Resource created" 
			            + (t == 0 ? "" : " (born:" + t + ")") 
			            + ": '" + resource + "'.");
			}
		}
	    return resource;
	}
	
	private void updateToMaxsize() {
		cleanExpiredResource();
		if (maxSize != UNLIMITED) {
			int delta = free.size() + used.size() - maxSize;
			//Remove the free pool resource if availlable
			if (delta > 0) {
				for (int i = 0; i < delta && !free.isEmpty(); i++) {
				    int j = free.size() - 1;
					Object resource = free.remove(j);
					if (freettl > 0) {
					    timesOfFree.remove(j);
					}
					if (logger != null && logger.isLoggable(BasicLevel.INFO)) {
					    logger.log(BasicLevel.INFO, "Resource destroyed: '" + resource + "'.");
					}
					matchFactory.destroyResource(resource);
				}
			}
		}
	}


    // IMPLEMENTATION OF METHODS FROM THE BindingController INTERFACE //
    //----------------------------------------------------------------//

    public String[] listFc() {
        return new String[] {
			POOL_MATCH_FACTORY_BINDING, DEPENDENCY_GRAPH_BINDING };
    }

    public Object lookupFc(String clientItfName) {
        if (POOL_MATCH_FACTORY_BINDING.equals(clientItfName)) {
            return matchFactory;
		} else if (DEPENDENCY_GRAPH_BINDING.equals(clientItfName)) {
            return dg;
		} else {
            return null;
		}
    }

    public void bindFc(String clientItfName, Object serverItf) {
        if (POOL_MATCH_FACTORY_BINDING.equals(clientItfName)) {
            matchFactory = (PoolMatchFactory) serverItf;
            try {
                updateToMinsize(null);
            } catch (Exception e) {
            }
		} else if (DEPENDENCY_GRAPH_BINDING.equals(clientItfName)) {
            dg = (DependencyGraph) serverItf;
        } else if ("logger".equals(clientItfName)) {
			logger = (Logger) serverItf;
		}
    }

    public void unbindFc(String clientItfName) {
        if (POOL_MATCH_FACTORY_BINDING.equals(clientItfName)) {
            matchFactory = null;
		} else if (DEPENDENCY_GRAPH_BINDING.equals(clientItfName)) {
            dg = null;
		}
    }


    // IMPLEMENTATION OF METHODS FROM THE Pool INTERFACE //
    //---------------------------------------------------//

    /**
     * <b>getResource</b> is used to allocate a PoolResource from the Pool.
     * Some hints are passed in order to specialise the matching or creation
     * of PoolResource.
     * @param hints	Some properties to specialise the matching or the creation
     *				of PoolResource.
     * @return		The PoolResource allocated from the Pool.
     */
    public synchronized Object getResource(Object hints) throws PoolException {
		try {
			return getResource(hints, null);
		} catch (DeadLockException e) {
			//not possible because no user is specified
			throw new PoolException(e);
		}
	}
    public synchronized Object getResource(Object hints, Object user)
			throws PoolException, DeadLockException {
		if (!started) {
			started = true;
			updateToMaxsize();
		}
		//System.out.println("getResource: max=" + (maxSize == UNLIMITED ? "U" : "" + maxSize) + ", free=" + free.size() + ", used=" + used.size() +")");
        if (maxSize != UNLIMITED && used.size() >= maxSize) {
            if (timeOut != 0) {
				int vertexIdx = 0;
				try {
				    //Add dependencies in order to detect dead lock
					if (user != null && dg != null) {
						int s = users.size();
						for(vertexIdx=0; vertexIdx<s; vertexIdx++) {
							Object dst = users.get(vertexIdx);
							if (dst != null) {
								if (!dg.addVertex(user, dst)) {
									throw new DeadLockException("Impossible to allocate a resource: no resource is availlable and a dead lock occurs if the current thread wait a resource");
								}
							}
						}
					}
					//wait a free resource or a max time
					if (timeOut > 0) {
						long to = timeOut;
						while(maxSize != UNLIMITED && used.size() >= maxSize && to > 0) {
							long exectime = System.currentTimeMillis();
							wait(timeOut);
							exectime = System.currentTimeMillis() - exectime;
							to -= exectime;
						}
					} else {
						while (maxSize != UNLIMITED && used.size() >= maxSize) {
							wait();
						}
					}
				} catch (InterruptedException e) {
					throw new PoolException("Pool wait has been interrupted: ", e);
				} finally {
					if (user != null && dg != null) {
						for(int i=0; i<vertexIdx; i++) {
							Object dst = users.get(i);
							if (dst != null) {
								dg.removeVertex(user, dst);
							}
						}
					}
				}
				if (maxSize != UNLIMITED && used.size() >= maxSize) {
                    throw new PoolException(
							"No more resource available in pool, time out exceeded ("
							+ timeOut + ") ! (pool max size=" + maxSize + ")");
                }
			} else {
				throw new PoolException("No more resource available in pool (max="
						+ maxSize + ")");
            }
        }
		updateToMinsize(hints);

		//search a free resource matching the hints
		Object resource = null;
        int freesize = free.size() -1;
        for (int i = freesize; resource == null && i >=0; i--) {
            resource = free.get(i);
            if (matchFactory.matchResource(resource, hints)) {
                free.remove(i);
				if (freettl > 0) {
                	timesOfFree.remove(i);
				}
				if (logger != null && logger.isLoggable(BasicLevel.DEBUG)) {
				    logger.log(BasicLevel.DEBUG, "Resource reused: '" + resource + "'.");
				}
			} else {
				resource = null;
			}
        }
        
	    if (resource == null) {
	        //no free and matching resource found ==> create it
	        resource = createResource(hints);
	    }

	    used.add(resource);
	    
	    //register the user for the dependencies management
		if (dg != null && user != null) {
		    users.add(user);
		    if (logger != null && logger.isLoggable(BasicLevel.DEBUG)) {
				logger.log(BasicLevel.DEBUG, "The user '" + user
						+ "' gets the resource '" + resource +"'.");
			}
		}
		return resource;
    }

    public synchronized int getSize() {
        return free.size() + used.size();
    }
    
    public synchronized int getFreeResourceNumber() {
        return free.size();
    }
    public synchronized int getUsedResourceNumber() {
        return used.size();
    }
    public synchronized Collection getUsers() {
        return Collections.unmodifiableCollection(users);
    }
    
    /**
     * <b>releaseResource</b> releases a PoolResource in order to allow the
     * Pool to recycle this PoolResource.
     * @param resource	The PoolResource to be released.
     */
    public synchronized void releaseResource(Object resource) throws PoolException {
        //System.out.println("releaseResource: max=" + (maxSize == UNLIMITED ? "U" : "" + maxSize) + ", free=" + free.size() + ", used=" + used.size() +")");
        boolean found = false;
		for(int i=(used.size() -1); !found && i>=0; i--) {
			if (resource == used.get(i)) {
			    found = true;
				used.remove(i);
				if (dg != null) {
					Object user = users.remove(i);
					if (logger != null && logger.isLoggable(BasicLevel.DEBUG)) {
						if (user == null) {
							logger.log(BasicLevel.DEBUG, "The resource '"
									+ resource +"' is released.");
						} else {
							logger.log(BasicLevel.DEBUG, "The user '" + user
									+ "' releases the resource '" + resource +"'.");
						}
					}
				}
			}
		}
		if (!found) {
			throw new PoolException("Release an unused resource:" + resource);
		}
        boolean toremove = false;
        boolean notify = true;
        if (maxSize != UNLIMITED
				&& (free.size() + used.size()) > maxSize) {
            toremove = true;
            notify = false;
			if (logger != null && logger.isLoggable(BasicLevel.INFO)) {
			    logger.log(BasicLevel.INFO, "Released resource is destroyed (too many resources in the pool): '" + resource + "'.");
			}
        } else {
            if (ttl > 0) {
                long t = Calendar.getInstance().getTimeInMillis();
                t -= ((Long) res2ttl.get(resource)).longValue();
                if (t > ttl) {
					if (logger != null && logger.isLoggable(BasicLevel.INFO)) {
					    logger.log(BasicLevel.INFO, "Released resource is being destroyed (too old): '" + resource + "'.");
					}
					toremove = true;
					//notify because a slot is available
                }
            }
        }
        if (toremove) {
            //The resource is removed from the pool
			if (ttl>0) {
			    res2ttl.remove(resource);
			}
			matchFactory.destroyResource(resource);
        } else {
            //The resource is put into the list of free resources
            free.add(resource);
			if (freettl > 0) {
				timesOfFree.add(new Long(Calendar.getInstance().getTimeInMillis()));
			}
        }
        if (notify) {
            //signal that a resource has been free
            notify();
        }
    }


    // IMPLEMENTATION OF THE PoolAttributes INTERFACE //
    //------------------------------------------------//

    /**
     * <b>getTimeout</b> retrieves the timeout assigned to this Pool.
     * @return	The timeout currently assigned to this Pool.
     */
    public long getTimeout() {
        return timeOut;
    }

    /**
     * <b>getMinSize</b> retrieves the minimum size assigned to this Pool.
     * @return	The minimum size currently assigned to this Pool.
     */
    public int getMinSize() {
        return minSize;
    }

    /**
     * <b>getMaxSize</b> retrieves the maximum size assigned to this Pool.
     * @return	The maximum size currently assigned to this Pool.
     */
    public int getMaxSize() {
        return maxSize;
    }

    /**
     * <b>setTimeout</b> assigns a timeout to this Pool.
     * @param crto		The timeout to be assigned.
     */
    public void setTimeout(long crto) {
        timeOut = crto;
    }

    /**
     * <b>setMinSize</b> assigns a minimum size to this Pool.
     * @param minsize	The minimum size to be assigned.
     */
    public void setMinSize(int minsize) throws Exception {
        minSize = minsize;
		if (started) {
        	updateToMinsize(null);
		}
    }

    /**
     * <b>setMaxSize</b> assigns a maximum size to this Pool.
     * @param maxsize	The maximum size to be assigned.
     */
    public void setMaxSize(int maxsize) throws Exception {
        maxSize = maxsize;
		if (started) {
        	updateToMaxsize();
		}
    }

	public long getTTL() {
		return ttl;
	}
	
	public void setTTL(long ttl) {
		this.ttl = ttl;
	}

	public long getInactiveTTL() {
		return freettl;
	}

	public void setInactiveTTL(long fttl) {
		this.freettl = fttl;
	}
}
