/**
 * Copyright (C) 2001-2003 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package org.objectweb.perseus.pool;

import junit.framework.TestCase;
import org.objectweb.util.monolog.api.Logger;
import org.objectweb.util.monolog.api.LoggerFactory;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.Monolog;
import org.objectweb.perseus.pool.api.Pool;
import org.objectweb.perseus.pool.api.PoolAttributes;
import org.objectweb.perseus.pool.api.PoolMatchFactory;
import org.objectweb.perseus.pool.api.PoolException;
import org.objectweb.perseus.pool.lib.ArrayListPool;
import org.objectweb.perseus.dependency.api.DependencyGraph;
import org.objectweb.perseus.dependency.api.DeadLockException;
import org.objectweb.perseus.dependency.lib.BasicDependencyGraph;
import org.objectweb.fractal.api.control.BindingController;

/**
 *
 * @author S.Chassande-Barrioz
 */
public class TestPool extends TestCase implements PoolMatchFactory {

	static Logger logger;
	static Logger poolLogger;
	static LoggerFactory loggerFactory;

	private int resourceCpt = 0;
	public TestPool(String s) {
		super(s);
		if (logger == null) {
			synchronized(Monolog.class) {
				if (logger == null) {
					loggerFactory = Monolog.initialize();
					logger = loggerFactory.getLogger(getClass().getName());
					poolLogger = loggerFactory.getLogger(
							instanciatePool().getClass().getName());
				}
			}
		}
	}

	protected void tearDown() throws Exception {
		resourceCpt = 0;
	}

	/**
	 * <b>createResource</b> creates a new PoolResource.
	 * @param hints	The "properties" that the created PoolResource should
	 *				conform to.
	 * @return		The created PoolResource.
	 */
	public Object createResource(Object hints) throws PoolException {
		resourceCpt ++;
		return (hints == null ? new Integer(resourceCpt) : hints);
	}

	/**
	 * <b>matchResource</b> tests if a given resource of a Pool matches with
	 * the hints passed with the Pool getResource method.
	 * @param resource	The PoolResource to test its matching with some
	 *				"properties" specified by hints.
	 * @param hints	The "properties" that the PoolResource specified by pr
	 *				should match.
	 * @return		<b>true</b> if the pr PoolResource matches the hints
	 *				"properties".
	 */
	public boolean matchResource(Object resource, Object hints) {
		return true;
	}

	/**
	 * is called when a resource is going to be destroyed. This method permits
	 * to close physical resource for example.
	 *
	 * @param resource to initialize.
	 */
	public void destroyResource(Object resource) {
	}

	protected Pool instanciatePool() {
		return new ArrayListPool();
	}
	private Pool getPool(int min, int max, long inactivettl, long ttl, long timeout, DependencyGraph dg) {
		Pool pool = instanciatePool();
		try {
			BindingController bc = (BindingController) pool;
			bc.bindFc("logger", poolLogger);
			bc.bindFc("monolog-factory", loggerFactory);
			bc.bindFc("dependency-graph", dg);
			bc.bindFc("pool-match-factory", this);
			PoolAttributes pa = (PoolAttributes) pool;
			pa.setMinSize(min);
			pa.setMaxSize(max);
			pa.setTTL(ttl);
			pa.setInactiveTTL(inactivettl);
			pa.setTimeout(timeout);
		} catch (Exception e) {
			String msg = "Initialisation error: ";
			logger.log(BasicLevel.ERROR, msg, e);
			fail(msg + e.getMessage());
		}
		return pool;
	}
	public void testNoWait() throws PoolException {
		Pool pool = getPool(0, 1, 0, -1, 0, null);
		pool.getResource(new Integer(0));
		try {
			pool.getResource(new Integer(3));
			fail("");
		} catch (PoolException e) {
		}
	}
	public void testWaitUntilRelease() throws PoolException {
		final Pool pool = getPool(0, 1, 0, -1, -1, null);
		Object r = pool.getResource(new Integer(0));
		final Boolean[] b = new Boolean[1];
		b[0] = Boolean.FALSE;
		Thread t = new Thread(new Runnable () {
			public void run () {
				try {
					pool.getResource(new Integer(3));
					if (!b[0].booleanValue()) {
						fail("Resource obtained");
					}
				} catch (PoolException e) {
					String msg = "Allocation error: ";
					logger.log(BasicLevel.ERROR, msg, e);
					fail(msg + e.getMessage());
				}
			}});
		t.start();
		Thread.yield();
		try {
			t.join(500);
		} catch (InterruptedException e) {
		}
		assertTrue("Thread not blocked", t.isAlive());
		b[0] = Boolean.TRUE;
		pool.releaseResource(r);
		try {
			t.join(500);
		} catch (InterruptedException e) {
		}
		assertTrue("Thread blocked", !t.isAlive());

	}

	public void testMinSize() throws PoolException {
		Pool pool = getPool(5, 10, 0, -1, 500, null);
		pool.getResource(new Integer(2));
		assertEquals("Bad min", 5, resourceCpt);
	}
	public void testWait500() throws PoolException {
		final Pool pool = getPool(0, 1, 0, -1, 500, null);
		pool.getResource(new Integer(0));
		Thread t = new Thread(new Runnable () {
			public void run () {
				try {
					pool.getResource(new Integer(3));
					fail("Resource obtained");
				} catch (PoolException e) {
				}
			}});
		t.start();
		try {
			t.join(600);
		} catch (InterruptedException e) {
		}
		assertTrue("Thread blocked", !t.isAlive());
	}

	protected DependencyGraph instanciateDependencyGraph() {
		return new BasicDependencyGraph();
	}
	protected DependencyGraph getDependencyGraph() {
		DependencyGraph dg = instanciateDependencyGraph();
		try {
			BindingController bc = (BindingController) dg;
			bc.bindFc("logger", poolLogger);
			bc.bindFc("monolog-factory", loggerFactory);
		} catch (Exception e) {
			String msg = "DG Initialisation error: ";
			logger.log(BasicLevel.ERROR, msg, e);
			fail(msg + e.getMessage());
		}
		return dg;
	}
	public void testDG() throws PoolException, DeadLockException {
		final Pool pool = getPool(0, 2, 0, -1, -1, getDependencyGraph());
		final String[] users = new String[] {"user1", "user2"};
		final Object[] resources = new Object[] {"res1", "res2"};
		pool.getResource(resources[0], users[0]);
		pool.getResource(resources[1], users[1]);
		Thread t = new Thread(new Runnable () {
			public void run () {
				try {
					 pool.getResource(resources[1], users[0]);
				} catch (Exception e) {
				}
			}});
		t.start();
		Thread.yield();
		try {
			t.join(100);
		} catch (InterruptedException e) {
		}
		try {
			pool.getResource(resources[0], users[1]);
			fail("No DeadLockException");
		} catch (DeadLockException e) {
		}
		pool.releaseResource(resources[0]);
		pool.releaseResource(resources[1]);
		try {
			t.join(200);
		} catch (InterruptedException e) {
		}
		assertTrue("Thread blocked", !t.isAlive());
	}

	public void testInactiveTTL() throws PoolException, DeadLockException {
		final int InactiveTTL = 5 * 1000; //10s
	    logger.log(BasicLevel.INFO, "testInactiveTTL(" + InactiveTTL + ")");
	    Pool pool = getPool(5, 10, InactiveTTL, -1, 500, null);
	    //init resource from the pool
	    pool.releaseResource(pool.getResource(null));
	    
	    //fetch a resource reference
	    Object r1 = pool.getResource(new Integer(2));
		pool.releaseResource(r1);
		
		//fetch again the resource: the inactivettl should not be expired
		//Then the resouce should be the same
		Object r2 = pool.getResource(new Integer(2));
		assertTrue("Resources are different, r1=" + r1 + ", r2=" + r2, r1 == r2);
		pool.releaseResource(r2);
		
		//Wait 2 InactiveTTL 
		try {
            Thread.sleep(2 * InactiveTTL);
        } catch (InterruptedException e) {
            fail(e.getMessage());
        }

        //fetch again the resource: the inactivettl should be expired
		//Then the resouce should be different
		r2 = pool.getResource(new Integer(2));
		assertTrue("Same resource, r1=" + r1 + ", r2=" + r2, r1 != r2);
	}	

	public void testTTL1() throws PoolException, DeadLockException {
		final int TTL = 5 * 1000;
	    logger.log(BasicLevel.INFO, "testTTL1(" + TTL + ")");
	    Pool pool = getPool(5, 10, -1, TTL, 500, null);
	    //init resource from the pool
	    pool.releaseResource(pool.getResource(null));

	    //fetch a resource reference
	    Object r1 = pool.getResource(new Integer(2));
		pool.releaseResource(r1);
	    
		//fetch again the resource: the ttl should not be expired
		//Then the resouce should be the same
		Object r2 = pool.getResource(new Integer(2));
		assertTrue("Resources are different, r1=" + r1 + ", r2=" + r2, r1 == r2);
		pool.releaseResource(r2);

		try {
            Thread.sleep(2 * TTL);
        } catch (InterruptedException e) {
            fail(e.getMessage());
        }

        //fetch again the resource: the ttl should be expired
		//Then the resouce should be different
		r2 = pool.getResource(new Integer(2));
		assertTrue("Same resource, r1=" + r1 + ", r2=" + r2, r1 != r2);
	}	

	public void testTTL2() throws PoolException, DeadLockException {
		final int TTL = 5 * 1000;
	    logger.log(BasicLevel.INFO, "testTTL2(" + TTL + ")");
	    Pool pool = getPool(3, 10, -1, TTL, 500, null);
	    //init resource from the pool
	    pool.releaseResource(pool.getResource(null));

	    //fetch a resource reference
	    Object r1 = pool.getResource(new Integer(2));
		pool.releaseResource(r1);
	    
		//fetch again the resource: the ttl should not be expired
		//Then the resouce should be the same
		Object r2 = pool.getResource(new Integer(2));
		assertTrue("Resources are different, r1=" + r1 + ", r2=" + r2, r1 == r2);

		try {
            Thread.sleep(2 * TTL);
        } catch (InterruptedException e) {
            fail(e.getMessage());
        }
		pool.releaseResource(r2);

        //fetch again the resource: the ttl should be expired
		//Then the resouce should be different
		r2 = pool.getResource(new Integer(2));
		assertTrue("Same resource, r1=" + r1 + ", r2=" + r2, r1 != r2);
	}	

	protected void setUp() throws Exception {
		allocTime = 0;
		nballoc = 0;
	}

	private long allocTime = 0;
	private long nballoc = 0;
	public void testStress(final int nbThread,
						   final int poolMaxSize,
						   final int nbAllocResource,
						   final int sleep
						   ) {
		logger.log(BasicLevel.INFO, "* testStress: nbThread=" + nbThread
				+ ", poolMaxSize=" + poolMaxSize
				+ ", nbAllocResource=" + nbAllocResource
				+ ", sleep=" + sleep);
		final Pool pool = getPool(0, poolMaxSize, 0, -1, -1, null);
		final Thread[] t = new Thread[nbThread];
		for(int i=0; i<nbThread; i++) {
			final int threadId = i;
			t[threadId] = new Thread(new Runnable () {
			public void run () {
				for(int j=0; j<nbAllocResource; j++) {
					try {
						long exectime = System.currentTimeMillis();
						Object r = pool.getResource(new Integer(3));
						if (sleep > 0) {
							Thread.sleep(sleep);
						}
						pool.releaseResource(r);
						exectime = System.currentTimeMillis() - exectime;
						synchronized(t) {
							nballoc ++;
							allocTime += exectime;
						}
					} catch (PoolException e) {
						String msg = "Thread=" + threadId  + ", alloc=" + j + " error: ";
						logger.log(BasicLevel.ERROR, msg, e);
						fail(msg + e.getMessage());
					} catch (InterruptedException e) {
					}
				}
			}});
		}
		long exectime = System.currentTimeMillis();
		for(int threadId=0; threadId<nbThread; threadId++) {
			t[threadId].start();
		}
		for(int threadId=0; threadId<nbThread; threadId++) {
			try {
				t[threadId].join(100000);
			} catch (InterruptedException e) {
			}
		}
		exectime = System.currentTimeMillis() - exectime;
		for(int threadId=0; threadId<nbThread; threadId++) {
			assertTrue("Thread blocked", !t[threadId].isAlive());
		}
		assertTrue("Bad pool size", pool.getSize() <= poolMaxSize);
		int nbAllAllocation = nbThread * nbAllocResource;
		logger.log(BasicLevel.INFO, "\tExecution time: " + exectime
				+ "ms, nb allocation: " + nbAllAllocation);
		logger.log(BasicLevel.INFO, "\tAverage Allocation time: " +
				(allocTime / nballoc) + "ms (sleep=" + sleep + "), rate: "
				+ ((nbAllAllocation * 1000) /exectime) + " alloc/s");
	}

	public void testStressTh20Max20Alloc100Sleep10() {
		testStress(20, 20, 100, 10);
	}
	public void testStressTh20Max20Alloc100Sleep100() {
		testStress(20, 20, 100, 100);
	}
	public void testStressTh20Max10Alloc100Sleep100() {
		testStress(20, 10, 100, 100);
	}
	public void testStressTh200Max100Alloc500Sleep50() {
		testStress(200, 100, 500, 50);
	}
	public void testStressTh100Max100Alloc5000Sleep0() {
		testStress(100, 100, 5000, 0);
	}
}
